import inspect
import pickle
from os import makedirs as make_dir
from pathlib import Path
from shutil import rmtree as rm_rf

from locache.logging import info

RUNNING_PATH = Path('.').absolute()


class LocalCache:
    def __init__(self, func, dirname=None, auto_invalidate=True) -> None:
        self._func = func
        self._dirname = dirname

        if not self.code_file.exists():
            # function has not been cached before
            self._setup()

        elif self._is_stale and not auto_invalidate:
            # cached calls are no longer valid
            self.reset()

    @property
    def code_file(self):
        return self._cache_dir / 'code.py'

    @property
    def code_i_am_caching(self):
        return inspect.getsource(self._func)

    def _setup(self):
        dir = self._cache_dir
        assert not dir.exists()
        make_dir(dir)

        self.code_file.write_text(self.code_i_am_caching)

    @property
    def _is_stale(self):
        old_code = self.code_file.read_text()
        new_code = self.code_i_am_caching
        return old_code != new_code

    @property
    def _cache_dir(self):
        if self._dirname is not None:
            # directory of currently running script + name
            _dir = RUNNING_PATH / f"{self._dirname}.cache"
        else:
            # get the file that func is defined in
            func_file = Path(inspect.getsourcefile(self._func)).absolute()
            _dir = func_file.with_suffix(".cache")

        return _dir / self._func.__name__

    def reset(self):
        info(f"WARNING: Resetting {self._func.__name__}'s cache.")
        rm_rf(self._cache_dir)
        self._setup()

    def file_path(self, args, kwargs):
        signature = inspect.signature(self._func)
        params = signature.bind(*args, **kwargs)
        params.apply_defaults()

        filename = '_'.join(
            f'{k}={v}' for k, v in params.arguments.items()
        )
        return self._cache_dir / filename

    @classmethod
    def load_from_disk(cls, _file):
        with open(_file, "rb") as file:
            return pickle.load(file)

    @classmethod
    def save_to_disk(cls, obj, _file):
        with open(_file, "wb") as file:
            pickle.dump(obj, file)



import functools
import inspect
import os

from locache.LocalCache import LocalCache, RUNNING_PATH
from locache.logging import info, verbose


def persist(_func=None, *, name=None, auto_invalidate=True):
    '''
    decorator that caches the return value of a function to 
    permanent memory for a given set of args and kwargs
    '''

    def persist_decorator(func):
        cache = LocalCache(func, name, auto_invalidate)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _file = cache.file_path(args, kwargs)
            _relative_path = os.path.relpath(_file, RUNNING_PATH)

            if _file.exists():  # cache hit: load from disk
                info(
                    f"Using cached result for {func.__name__} "
                    f"from {_relative_path}"
                )
                out = LocalCache.load_from_disk(_file)

            else:  # cache miss: call func and save to disk
                info(
                    f"Caching result for {func.__name__} "
                    f"in {_relative_path}"
                )
                out = func(*args, **kwargs)
                LocalCache.save_to_disk(out, _file)

            return out

        return wrapper

    if _func == None:  # called as @persist(name=...)
        return persist_decorator
    else:              # called as @persist
        return persist_decorator(_func)


def reset_cache(func, name=None):
    cache = LocalCache(inspect.unwrap(func), name)
    cache.reset()



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="locache",
    version="2.0.2",
    description="Cache expensive function calls to disk.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jla-gardner/local-cache",
    author="John Gardner",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["locache"],
    keywords=["cache"],
    install_requires=[],
    package_data={},
    python_requires=">=3.7, <4",
)
